﻿//
// Copyright (c) 2020, MindFusion LLC - Bulgaria.
//

using MindFusion.Diagramming.Wpf.Lanes;
using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using Grid = MindFusion.Diagramming.Wpf.Lanes.Grid;


namespace MindFusion.Diagramming.Wpf.Samples.CS.Lanes
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : Window
    {
        public MainWindow()
        {
            InitializeComponent();

            // Initialize flowchart
			diagram.AutoResize = AutoResize.None;
			diagram.AllowInplaceEdit = true;

			//custom anchor pattern for nodes
            pattern = new AnchorPattern(new[] {
				new AnchorPoint(100, 50, false, true, MarkStyle.None),
				new AnchorPoint(0, 0, true, false, MarkStyle.None)
			});
			//add default style for diagram nodes
			var style = new System.Windows.Style();
			style.Setters.Add(new Setter(DiagramNode.BrushProperty, new SolidColorBrush(Color.FromRgb(237, 237, 237))));		
			diagram.ShapeNodeStyle = style;

			style = new System.Windows.Style();
			style.Setters.Add(new Setter(DiagramLink.BrushProperty, Brushes.Transparent));
			style.Setters.Add(new Setter(DiagramLink.StrokeProperty, Brushes.Transparent));
			diagram.DiagramLinkStyle = style;
			

			//add event handler for the DrawLink event
			diagram.DrawLink += Diagram_DrawLink;
			//handler for the link creating event
			diagram.LinkCreating += Diagram_LinkCreating;
	
			

            //Initialize grid
            Header header;
            Grid grid = diagram.LaneGrid;

            HeaderCollection columns = grid.ColumnHeaders;
			//grid settings
            grid.HeadersOnTop = false;
			grid.ColumnHeadersHeights = new double[] { 5 };
            grid.AlignCells = false;
            grid.AllowResizeHeaders = true;
			
			//create column headers
			for (int i = 0; i < 8; i++)
            {
                header = new Header(i.ToString());
				header.Width = 140;
				header.Style.BackgroundBrush = Brushes.Transparent;
				header.Style.LeftBorderPen = header.Style.TopBorderPen = header.Style.BottomBorderPen =
					header.Style.RightBorderPen = new Pen(Brushes.Transparent, 1.0);
				columns.Add(header);
			}
					   
			//titles for grid rows
			string[] lanes = {"Action", "", "Confirmation", "Interaction", "Follow Up", "Preparation" };
			//heights of grid lanes
			double[] heights = { 100, 30, 110, 140, 220, 180 };

			//create the row headers
			for (int i = 1; i <= 6; i++)
            {
				header = new Header(lanes[i-1]);
				header.FontWeight = FontWeights.Bold;
			    header.Width = 140;
				header.Height = heights[i-1];
				Wpf.Lanes.Style hstyle = new Wpf.Lanes.Style();
				if((i % 5) == 0)
					hstyle.BackgroundBrush = new SolidColorBrush(Color.FromRgb(237, 237, 237));
				if ((i % 6) == 0)
					hstyle.BackgroundBrush = new SolidColorBrush(Color.FromRgb(220, 220, 220));

				hstyle.BottomBorderPen = hstyle.LeftBorderPen =
					hstyle.RightBorderPen = hstyle.TopBorderPen = 
					new Pen(hstyle.BackgroundBrush, 1.0);

				header.Style = hstyle;
				grid.RowHeaders.Add(header);
			}


			ShapeNode shape;
			Rect bounds;

			//the first lane with the images
			string[] file_names = { "website", "registration", "mails", "gifts", "updates", "details", "coupon", "message" };

			for(int i = 0; i < 8; i++)
			{
				//get the bounds of this grid cell
				bounds = grid.GetCellBounds(i, 0);
				Rect imageRect = new Rect(bounds.X, bounds.Y + 10, bounds.Width - 20, bounds.Height - 20);
				//the shape is transparent
				shape = diagram.Factory.CreateShapeNode(imageRect, Shapes.Rectangle);
				shape.Transparent = true;	
				shape.Image = new BitmapImage(new Uri("pack://application:,,,/Lanes;component/Images/" + file_names[i] + ".png"));
				//the tag shows the lane index
				shape.Tag = 0;

			}

			//second lane with labels under the images
			string[] labels = { "Website\nvisit", "Register", "Confirmation\ne-mail", "Surprise\ngift", "Website\nUpdates",
				"Customer\ndetails", "Purcahse\ncoupons", "Company\nMessage" };

			//the cells contain rectangles that render under the images
			for(int i = 0; i < 8; i++)
			{
				bounds = grid.GetCellBounds(i, 1);
				bounds.Offset(-10, 0);
				shape = diagram.Factory.CreateShapeNode(bounds, Shapes.Rectangle);
				shape.StrokeThickness = 0L;
				shape.Text = labels[i];
				shape.FontSize = 11;
				shape.Transparent = true;
				//the tag shows the lane index
				shape.Tag = 1;				
			}


			/* set the background of the cells in each lane. Take the color
			 * from the background of the row header */

			for (int i = 0; i < 6; i++)
			{
				Header row = grid.GetRow(i);

				grid[null, row].Style.BackgroundBrush = row.Style.BackgroundBrush;
				grid[null, row].Style.BottomBorderPen = grid[null, row].Style.LeftBorderPen =
				grid[null, row].Style.RightBorderPen = grid[null, row].Style.TopBorderPen =
				new Pen(row.Style.BackgroundBrush, 1.0);
				if (i < 2)
				{
					grid[null, row].Style.BackgroundBrush = new SolidColorBrush(Color.FromRgb(237, 237, 237));
					grid[null, row].Style.LeftBorderPen = grid[null, row].Style.TopBorderPen =
						grid[null, row].Style.BottomBorderPen = new Pen(new SolidColorBrush(Color.FromRgb(237, 237, 237)), 1.0);
					grid[null, row].Style.RightBorderPen = new Pen(Brushes.White, 40.0);
				}
				else if (i < 5)
					grid[null, row].Style.BottomBorderPen = new Pen(new SolidColorBrush(Color.FromRgb(100, 100, 100)), 3.5);
			}

			//third lane - dark pink arrows
			string[] confirmation = {"Finds our\nwebsite", "Fills in the\nregistration form", "Reads the\nconfirmation mail", "Receives\nsurprise gifts",
				"Web updates", "Collect\nuser data", "Give coupons", "Understands\nour message"};

			for (int i = 0; i < confirmation.Length; i++)
			{
				bounds = grid.GetCellBounds(i, 2);
				Rect nodeRect = new Rect(bounds.X + 5, bounds.Y + 25, bounds.Width - 10, bounds.Height - 50);

				//the shape of the first shape is different
				if (i == 0)
					shape = diagram.Factory.CreateShapeNode(nodeRect, Shapes.Arrow7);
				else
					shape = diagram.Factory.CreateShapeNode(nodeRect, Shapes.IsoProcess);
				shape.Text = confirmation[i];
				//the id is the same as the text
				shape.Id = confirmation[i];
				//the tag shows the lane index
				shape.Tag = 2;
				shape.FontSize = 10;
				shape.Brush = new SolidColorBrush(Color.FromRgb(243, 121, 114));
				shape.Stroke = new SolidColorBrush(Color.FromRgb(241, 98, 91));
				shape.TextBrush = Brushes.White;
				
			}

			//fourth lane - rectangles with text
			string[] interaction = {"Social networks,\ndigital ads,\nbrochures", "Joins our netowrk",
				"Personalized\noffers", "Checks regularly\nfor new info", "Gathers the coupons", "Makes purchases",
			"Invitations to\njoin clubs"};

			//we keep the nodes in this lane in a list
			List<ShapeNode> laneNodes = new List<ShapeNode>(interaction.Length);

			for (int i = 0; i < interaction.Length; i++)
			{
				bounds = grid.GetCellBounds(i, 3);

				Rect nodeRect = new Rect(bounds.X, bounds.Y + 20, bounds.Width - 10, bounds.Height - 20);

				shape = diagram.Factory.CreateShapeNode(nodeRect, Shapes.Rectangle);
				//the tag shows the lane index
				shape.Tag = 3;
				shape.Text = interaction[i];
				shape.Id = interaction[i];
				shape.FontSize = 10;
				shape.Brush = new SolidColorBrush(Color.FromRgb(238, 62, 54));
				shape.Stroke = new SolidColorBrush(Color.FromRgb(235, 31, 20));
				shape.TextBrush = Brushes.White;
				shape.TextPadding = new Thickness(10L);
				shape.ResizeToFitText(FitSize.KeepWidth);
				laneNodes.Add(shape);
			}

			//we create the links between nodes from the last two lanes
			DiagramNode node1 = diagram.FindNodeById("Finds our\nwebsite");
			DiagramNode node2 = diagram.FindNodeById("Social networks,\ndigital ads,\nbrochures");

			if(node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);


			node1 = diagram.FindNodeById("Fills in the\nregistration form");
			node2 = diagram.FindNodeById("Joins our netowrk");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Reads the\nconfirmation mail");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Receives\nsurprise gifts");
			node2 = diagram.FindNodeById("Personalized\noffers");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);
			
			node1 = diagram.FindNodeById("Invitations to\njoin clubs");
			if (node1 != null)
				diagram.Factory.CreateDiagramLink(node2, node1);

			node1 = diagram.FindNodeById("Web updates");
			node2 = diagram.FindNodeById("Checks regularly\nfor new info");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Collect\nuser data");
			if (node1 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node2 = diagram.FindNodeById("Gathers the coupons");
			if (node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Give coupons");
			if (node1 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Understands\nour message");
			node2 = diagram.FindNodeById("Makes purchases");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			//we arrange the nodes from this lane depending on their links
			arrangeNodes(laneNodes);

			//labels for the nodes for the fourth lane
			string[] follow_up = {"Personal posts,\nchats, likes,\ncomments, photos", "Do they log\nregularly?",
				 "Give pamphlets,\nmagazines", "Send more coupons","Send 'thank you'\ne-mail", "Give loyalty\npoints", "Send prompting\nemail",
				 "Send new\nbonus\nproducts", "Offer free\nguidance and tips","Offer\n assistance"};

			laneNodes.Clear();

			for (int i = 0; i < follow_up.Length; i++)
			{
				//we get the bounds of the cell
				bounds = grid.GetCellBounds(Math.Min(i, 7), 4);

				Rect nodeRect = new Rect(bounds.X, bounds.Y + 20, bounds.Width - 40, bounds.Height - 20);

				if(i == follow_up.Length - 1)
					shape = diagram.Factory.CreateShapeNode(nodeRect, Shapes.Decision);
				else
					shape = diagram.Factory.CreateShapeNode(nodeRect, Shapes.Rectangle);
				//the tag shows the lane index
				shape.Tag = 4;
				shape.Text = follow_up[i];
				shape.Id = follow_up[i];
				shape.FontSize = 10;
				shape.Brush = new SolidColorBrush(Color.FromRgb(153,153,153));
				shape.Stroke = new SolidColorBrush(Color.FromRgb(120,120,120));
				shape.TextBrush = Brushes.White;
				shape.TextPadding = new Thickness(10L);
				shape.ResizeToFitText(FitSize.KeepWidth);
				laneNodes.Add(shape);
			}

			//we create links among these nodes and those of the previous lanes
			node1 = diagram.FindNodeById("Social networks,\ndigital ads,\nbrochures");
			node2 = diagram.FindNodeById("Personal posts,\nchats, likes,\ncomments, photos");		

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Joins our netowrk");
			node2 = diagram.FindNodeById("Do they log\nregularly?");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Invitations to\njoin clubs");
			node2 = diagram.FindNodeById("Give pamphlets,\nmagazines");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Gathers the coupons");
			node2 = diagram.FindNodeById("Send more coupons");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Makes purchases");
			node2 = diagram.FindNodeById("Send 'thank you'\ne-mail");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Give loyalty\npoints");
			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node2, node1);


			node1 = diagram.FindNodeById("Do they log\nregularly?");
			node2 = diagram.FindNodeById("Send prompting\nemail");

			//this link has a label
			if (node1 != null && node2 != null)
			{
				((ShapeNode)node2).ResizeToFitText(FitSize.KeepHeight);			
				DiagramLink link = diagram.Factory.CreateDiagramLink(node1, node2);
				link.Text = "No";
			}

			node1 = diagram.FindNodeById("Offer\n assistance");

			if (node1 != null)
				diagram.Factory.CreateDiagramLink(node2, node1);

			node1 = diagram.FindNodeById("Give pamphlets,\nmagazines");
			node2 = diagram.FindNodeById("Send new\nbonus\nproducts");

			if (node1 != null && node2 != null)
			{
				((ShapeNode)node2).ResizeToFitText(FitSize.KeepHeight);
				DiagramLink link = diagram.Factory.CreateDiagramLink(node1, node2);				
			}

			node1 = diagram.FindNodeById("Send more coupons");
			node2 = diagram.FindNodeById("Offer free\nguidance and tips");
			if (node1 != null && node2 != null)						
				diagram.Factory.CreateDiagramLink(node1, node2);

			//we arrange the nodes from this labe
			arrangeNodes(laneNodes);

			string[] preparation = { "Create appealing,\neasy to use website",
				"Make sure the interface\nis user friendly",
				"Maintain the\nregistration portal\nand the website,\nprioritize security",
				"Publish all discounts online", "Calculate\ncoupon price",
				"Maintain\appealing\nproduct portfolio" };


			laneNodes.Clear();

			//we create the nodes for the last lane and add them to the nodes list
			for (int i = 0; i < preparation.Length; i++)
			{
				bounds = grid.GetCellBounds(Math.Min(i, 5), 5);

				Rect nodeRect = new Rect(bounds.X, bounds.Y + 20, bounds.Width - 40, bounds.Height - 20);
				shape = diagram.Factory.CreateShapeNode(nodeRect, Shapes.Rectangle);
				//the tag shows the lane index
				shape.Tag = 5;
				shape.Text = preparation[i];
				shape.Id = preparation[i];
				shape.FontSize = 10;
				shape.Brush = new SolidColorBrush(Color.FromRgb(153, 153, 153));
				shape.Stroke = new SolidColorBrush(Color.FromRgb(120, 120, 120));
				shape.TextBrush = Brushes.White;
				shape.TextPadding = new Thickness(10L);
				shape.ResizeToFitText(FitSize.KeepWidth);
				laneNodes.Add(shape);
			}

			//we create the links between nodes from this lane and nodes from previous lanes
			node1 = diagram.FindNodeById("Personal posts,\nchats, likes,\ncomments, photos");
			node2 = diagram.FindNodeById("Create appealing,\neasy to use website");

			if (node1 != null && node2 != null)			
				 diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Offer\n assistance");
			node2 = diagram.FindNodeById("Make sure the interface\nis user friendly");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Do they log\nregularly?");
			if (node1 != null && node2 != null)
			{		
				DiagramLink link = diagram.Factory.CreateDiagramLink(node1, node2);
				link.Text = "Yes";
			}

			node1 = diagram.FindNodeById("Maintain the\nregistration portal\nand the website,\nprioritize security");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node2, node1);

			node1 = diagram.FindNodeById("Invitations to\njoin clubs");
			node2 = diagram.FindNodeById("Publish all discounts online");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Calculate\ncoupon price");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node2, node1);

			node2 = diagram.FindNodeById("Offer free\nguidance and tips");
			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node2, node1);


			node2 = diagram.FindNodeById("Publish all discounts online");
			node1 = diagram.FindNodeById("Checks regularly\nfor new info");
			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			node1 = diagram.FindNodeById("Give loyalty\npoints");
			node2 = diagram.FindNodeById("Maintain\appealing\nproduct portfolio");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);

			//we arrange the nodes
			arrangeNodes(laneNodes);

			node1 = diagram.FindNodeById("Joins our netowrk");
			node2 = diagram.FindNodeById("Maintain the\nregistration portal\nand the website,\nprioritize security");

			if (node1 != null && node2 != null)
				diagram.Factory.CreateDiagramLink(node1, node2);
			//make node1 bigger
			var rect = node1.Bounds;
			rect.Inflate(20, 0);
			rect.Offset(20, 0);
			node1.Bounds = rect;

			ArrangeLinks();
			//turn on lanels for the diagram
			diagram.EnableLanes = true;

			//route the diagram links
			Layout.OrthogonalRouter router = new Layout.OrthogonalRouter();
			router.Arrange(diagram);
			
			//set full custom drawing for all links
			//make the links be drawn after nodes
			foreach(DiagramLink link in diagram.Links)
			{
				link.CustomDraw = CustomDraw.Full;
				link.ZIndex = 1;
			}

			//assign the custom anchor pattern to all nodes
			foreach (ShapeNode node in diagram.Nodes)
				node.AnchorPattern = pattern;
			//turn off alignment to grid
             diagram.AlignToGrid = false;
        }

		private void Diagram_InitializeLink(object sender, LinkEventArgs e)
		{
			e.Link.CustomDraw = CustomDraw.Full;
		}

		private void Diagram_LinkCreating(object sender, LinkValidationEventArgs e)
		{
			e.Link.CustomDraw = CustomDraw.Full;
		}
	

		//custom drawing of diagram links
		private void Diagram_DrawLink(object sender, DrawLinkEventArgs e)
		{
			//brushes for the links
			Brush[] laneBrushes = { Brushes.Black,
				Brushes.Black,
				 new SolidColorBrush(Color.FromRgb(243, 121, 114)),
				 new SolidColorBrush(Color.FromRgb(238, 62, 54)),
				 new SolidColorBrush(Color.FromRgb(207, 207, 207)),
				 new SolidColorBrush(Color.FromRgb(170, 170, 170)) };

			//get a collection of the points for each link as divided by segments
			PointCollection points = getLinkSegments(e.Points);
			double width = 10.0;
			double halfWidth = width / 2;

			for(int i = 0; i < points.Count - 1; i++)
			{
				int index = pointInSegment(points[i]);
				Rect drawingRect = Rect.Empty;
				if (points[i].X == points[i + 1].X)//the segment is vertical
					drawingRect = new Rect(
						points[i].X - halfWidth, points[i].Y - halfWidth,
						width, Math.Abs(points[i + 1].Y - points[i].Y) + width);
				else if (points[i].X < points[i + 1].X)//the segment is horizontal, left to right
					drawingRect = new Rect(
						points[i].X - halfWidth, points[i].Y - halfWidth,
						points[i + 1].X - points[i].X + width, width);
				else//points[i].x > points[i+1].x the segment is horizontal, right to left
					drawingRect = new Rect(
						points[i + 1].X - halfWidth, points[i].Y - halfWidth,
						points[i].X - points[i + 1].X + width, width);

				e.Graphics.DrawRectangle(laneBrushes[index], null, drawingRect);
			}
		}


		//divide the link into segments that depend on its crossing of diagram lanes
		private PointCollection getLinkSegments(PointCollection points)
		{
			Grid grid = diagram.LaneGrid;
			PointCollection segments = new PointCollection(0);

			//get the first and last lane of the link
			int firstSegment = pointInSegment(points[0]);
			int lastSegment = pointInSegment(points[points.Count - 1]);

			double currPos = points[0].Y;
			int counter = pointInSegment(points[0]);

			for (int i = firstSegment; i <= lastSegment; i++)
			{
				double segmentStart = grid.GetCellBounds(0, i).Top;
				double segmentEnd = grid.GetCellBounds(0, i).Bottom;

				//check if the link point is inside the lane
				for (int j = 0; j < points.Count; j++)
				{
					if (points[j].Y >= segmentStart && points[j].Y < segmentEnd)
						segments.Add(points[j]);
				}

				//if the point is not inside the lane, we add the lane border as link point
				if (i < lastSegment)
					segments.Add(new Point(segments[segments.Count - 1].X, segmentEnd));
			}


			return segments;
		}
			
	
		//find the segment, which contains the point
		private int pointInSegment(Point point)
		{
			Grid grid = diagram.LaneGrid;

			for (int i = 0; i < grid.RowCount; i++)
			{
				double segmentStart = grid.GetCellBounds(0, i).Top;
				double segmentEnd = grid.GetCellBounds(0, i).Bottom;

				if (point.Y >= segmentStart && point.Y < segmentEnd)
					return i;
			}

			return -1;
		}


		// distributes message links between processes so that they are horizontal and
		// do not overlap when there are more than one links connecting the same pair
		private void ArrangeLinks()
		{
			var bundles = FindLinkBundles();
			foreach (var bundle in bundles.Values)
			{
				var range = SharedRange(bundle[0].Origin, bundle[0].Destination);
				if (range != null)
				{
					double h = range[1] - range[0];
					double s = h / (bundle.Count + 1);
					double x = range[0] + s;
					foreach (DiagramLink link in bundle)
					{
						link.Shape = LinkShape.Polyline;
						link.SegmentCount = 1;
						link.ReassignAnchorPoints();

						link.StartPoint = new Point(x, link.StartPoint.Y);
						link.EndPoint = new Point(x, link.EndPoint.Y);
						link.UpdateFromPoints();

						x += s;
					}
				}
				else
				{
					diagram.RoutingOptions.StartOrientation = Orientation.Horizontal;
					diagram.RoutingOptions.EndOrientation = Orientation.Vertical;

					DiagramLink l = bundle[0];
					DiagramNode left, right;
					if (l.Origin.Bounds.X < l.Destination.Bounds.X)
					{
						left = l.Origin;
						right = l.Destination;
					}
					else
					{
						left = l.Destination;
						right = l.Origin;
					}

					double ly, ry, h = 20, s = h / (bundle.Count + 1);
					if (left.Bounds.Bottom < right.Bounds.Top)
					{
						ly = s + left.Bounds.Bottom - h;
						ry = s + right.Bounds.Top;
					}
					else
					{
						ly = s + left.Bounds.Top;
						ry = s + right.Bounds.Bottom - h;
					}
					foreach (DiagramLink link in bundle)
					{
						link.Shape = LinkShape.Cascading;
						if (link.Origin == left)
						{
							link.StartPoint = new Point(left.Bounds.Right, ly);
							link.EndPoint = new Point(right.Bounds.Left, ry);
						}
						else
						{
							link.StartPoint = new Point(right.Bounds.Left, ry);
							link.EndPoint = new Point(left.Bounds.Right, ly);
						}
						ly += s;
						ry += s;
					}
					diagram.LinkRouter.RouteLinks(bundle);
				}
			}
		}

		private Dictionary<int, DiagramLinkCollection> FindLinkBundles()
		{
			var bundles = new Dictionary<int, DiagramLinkCollection>();
			foreach (DiagramLink link in diagram.Links)
			{
				if ("Lifeline".Equals(link.Tag))
					continue;

				int code = PairCode(link);
				if (!bundles.ContainsKey(code))
					bundles[code] = new DiagramLinkCollection();
				bundles[code].Add(link);
			}
			return bundles;
		}

		private double[] SharedRange(DiagramNode n1, DiagramNode n2)
		{
			var r1 = n1.Bounds;
			var r2 = n2.Bounds;

			if (r1.Right > r2.Left)
				return null;
			if (r1.Left < r2.Right)
				return null;

			return new[] { Math.Min(r1.Left, r2.Left), Math.Max(r1.Right, r2.Right) };
		}

		private int PairCode(DiagramLink link)
		{
			int i1 = link.Origin.ZIndex;
			int i2 = link.Destination.ZIndex;
			return diagram.Items.Count * Math.Max(i1, i2) + Math.Min(i1, i2);
		}

		private List<ShapeNode> arrangeNodes(List<ShapeNode> laneNodes)
		{
			//get all diagram nodes with 2 incoming links
			for (int i = 0; i < laneNodes.Count; i++)
			{
				ShapeNode n = laneNodes[i];
				Rect originRect = n.Bounds;

				if (n.IncomingLinks.Count > 1)
				{
					double middle = 0;
					foreach (DiagramLink link in n.IncomingLinks)
					{
						Rect nbounds = link.Origin.Bounds;
						double center = nbounds.Left + nbounds.Width / 2;
						middle += center;
					}

					middle /= n.IncomingLinks.Count;
					n.Bounds = new Rect(middle - originRect.Width / 2, originRect.Y, originRect.Width, originRect.Height);
				}
				else if (n.IncomingLinks.Count == 1)
				{					
					DiagramLink link = n.IncomingLinks[0];
					Rect nbounds = link.Origin.Bounds;
					double yPosition = originRect.Y;

					if (link.Origin.Tag.ToString() == n.Tag.ToString())
					{
						yPosition = nbounds.Bottom + 20;
					}

										
					n.Bounds = new Rect(nbounds.X, yPosition, originRect.Width, originRect.Height);

				}
			}

			return laneNodes;
		}

		private AnchorPattern pattern;

        /// <summary>
        /// Fixes the segments of the specified link according
        /// to the position of its related boxes.
        /// </summary>
        private static void FixLink(DiagramLink link)
        {
            Rect originBounds = link.Origin.Bounds;
            Rect destBounds = link.Destination.Bounds;

            if (originBounds.Right < destBounds.Left)
            {
                link.SegmentCount = 2;
            }
            else
            {
                link.SegmentCount = 4;

                Point pt0 = link.ControlPoints[0];
                Point pt1;
                Point pt2;
                Point pt3 = link.ControlPoints[3];
                Point pt4 = link.ControlPoints[4];

                pt1 = pt0;
                pt1.X += 6;

                pt2 = pt1;
                pt2.Y += 15;

                pt3.X = pt4.X;
                pt3.Y = pt2.Y;

                link.ControlPoints[1] = pt1;
                link.ControlPoints[2] = pt2;
                link.ControlPoints[3] = pt3;

                link.UpdateFromPoints();
            }
        }

        private void diagram_HitTestAdjustmentHandles(object sender, HitTestEventArgs e)
        {
            if (!(e.Item is ShapeNode))
            {
                e.HitResult = null;
                return;
            }

            var node = e.Item as ShapeNode;

            if (!node.Bounds.Contains(e.MousePosition))
            {
				e.HitResult = null;
                return;
            }

            Rect left = node.Bounds;
            left.Width = 2;

            if (left.Contains(e.MousePosition))
            {
				e.HitResult = new NodeAdjustmentHandle(NodeHandleType.ResizeMiddleLeft);
                return;
            }

            Rect right = node.Bounds;
            right.X = right.Right - 2;
            right.Width = 2;

            if (right.Contains(e.MousePosition))
            {
				e.HitResult = new NodeAdjustmentHandle(NodeHandleType.ResizeMiddleRight);
                return;
            }

            Rect middle = node.Bounds;
            middle.X = middle.X + middle.Width / 4;
            middle.Width = middle.Width / 2;

            if (middle.Contains(e.MousePosition))
            {
				e.HitResult = new NodeAdjustmentHandle(NodeHandleType.Move);
                return;
            }

			e.HitResult = null;
        }

        private void diagram_NodeCreated(object sender, NodeEventArgs e)
        {
          //  e.Node.HandlesStyle = HandlesStyle.Custom;
            e.Node.AnchorPattern = pattern;

			Brush[] nodeBrushes = {  new SolidColorBrush(Color.FromRgb(237, 237, 237)),
				new SolidColorBrush(Color.FromRgb(237, 237, 237)),
				 new SolidColorBrush(Color.FromRgb(243, 121, 114)),
				 new SolidColorBrush(Color.FromRgb(238, 62, 54)),
				 new SolidColorBrush(Color.FromRgb(153, 153, 153)),
				 new SolidColorBrush(Color.FromRgb(153, 153, 153)) };

			Brush[] nodeStrokes = {  new SolidColorBrush(Color.FromRgb(207, 207, 207)),
				new SolidColorBrush(Color.FromRgb(207, 207, 207)),
				 new SolidColorBrush(Color.FromRgb(241, 98, 91)),
				 new SolidColorBrush(Color.FromRgb(235, 31, 20)),
				 new SolidColorBrush(Color.FromRgb(120, 120, 120)),
				 new SolidColorBrush(Color.FromRgb(120, 120, 120)) };

			// Place the box within the grid
			Rect bounds = e.Node.Bounds;
            Point topLeft = new Point(bounds.Left, bounds.Top);

            Grid grid = diagram.LaneGrid;

            Rect cellBounds = Rect.Empty;
            grid.GetCellFromPoint(topLeft, ref cellBounds);

			int laneIndex = pointInSegment(cellBounds.TopLeft);

			if (laneIndex == 2)
				((ShapeNode)e.Node).Shape = Shapes.IsoProcess;
			else
				((ShapeNode)e.Node).Shape = Shapes.Rectangle;

			e.Node.Brush = nodeBrushes[laneIndex];
			e.Node.Stroke = nodeStrokes[laneIndex];
			e.Node.TextBrush = Brushes.White;
			e.Node.FontSize = 10;

		}

        private void diagram_NodeCreating(object sender, NodeValidationEventArgs e)
        {
			Brush[] nodeBrushes = {  new SolidColorBrush(Color.FromRgb(237, 237, 237)),
				new SolidColorBrush(Color.FromRgb(237, 237, 237)),
				 new SolidColorBrush(Color.FromRgb(243, 121, 114)),
				 new SolidColorBrush(Color.FromRgb(238, 62, 54)),
				 new SolidColorBrush(Color.FromRgb(153, 153, 153)),
				 new SolidColorBrush(Color.FromRgb(153, 153, 153)) };

			Brush[] nodeStrokes = {  new SolidColorBrush(Color.FromRgb(207, 207, 207)),
				new SolidColorBrush(Color.FromRgb(207, 207, 207)),
				 new SolidColorBrush(Color.FromRgb(241, 98, 91)),
				 new SolidColorBrush(Color.FromRgb(235, 31, 20)),
				 new SolidColorBrush(Color.FromRgb(120, 120, 120)),
				 new SolidColorBrush(Color.FromRgb(120, 120, 120)) };

			Rect bounds = e.Node.Bounds;
            Point topLeft = new Point(bounds.Left, bounds.Top);

            Grid grid = diagram.LaneGrid;

            Rect cellBounds = Rect.Empty;
            ICell cell = grid.GetCellFromPoint(topLeft, ref cellBounds);

            if (cell == null)
                e.Cancel = true;			

			int laneIndex = pointInSegment(cellBounds.TopLeft);

			if (laneIndex == 2)
				((ShapeNode)e.Node).Shape = Shapes.IsoProcess;
			else
				((ShapeNode)e.Node).Shape = Shapes.Rectangle;

			e.Node.Brush = nodeBrushes[laneIndex];
			e.Node.Stroke = nodeStrokes[laneIndex];
			e.Node.TextBrush = Brushes.White;
			e.Node.FontSize = 10;
		}     

        private void diagram_NodeModifying(object sender, NodeValidationEventArgs e)
        {
            Rect bounds = e.Node.Bounds;
            Point topLeft = new Point(bounds.Left, bounds.Top + bounds.Height / 2);

            Grid grid = diagram.LaneGrid;

            Rect cellBounds = Rect.Empty;
            ICell cell = grid.GetCellFromPoint(topLeft, ref cellBounds);

            if (cell == null || cellBounds.IsEmpty)
                e.Cancel = true;
        }

		private void diagram_SelectionModifying(object sender, SelectionValidationEventArgs e)
		{
			Rect bounds = diagram.Selection.Bounds;

			var gridBounds = diagram.LaneGrid.GetCellBounds(diagram.LaneGrid[null, null]);
			if (!gridBounds.Contains(bounds))
				e.Cancel = true;
		}

        private void diagram_LinkCreated(object sender, LinkEventArgs e)
        {
            e.Link.AllowMoveStart = false;
            e.Link.AllowMoveEnd = false;
            e.Link.HandlesStyle = HandlesStyle.Custom;
			e.Link.CustomDraw = CustomDraw.Full;

            FixLink(e.Link);
        }

        private void diagram_NodeModified(object sender, NodeEventArgs e)
        {
            // Place the box within the grid
            Rect bounds = e.Node.Bounds;
            Point topLeft = new Point(bounds.Left, bounds.Top + bounds.Height / 2);

            Grid grid = diagram.LaneGrid;

            Rect cellBounds = Rect.Empty;
            grid.GetCellFromPoint(topLeft, ref cellBounds);
			if (cellBounds.IsEmpty)
				return;      

            // Fix links
        	var links = e.Node.GetAllLinks();
            foreach (DiagramLink link in links)
                FixLink(link);
        }
    }
}